#pragma once

/* The Sitara VIM requires IRQ handler addresses to be 4-byte aligned, even if
 * they are thumb-encoded. (The least-significant 2 bits are ignored.) */
#define RT_VIC_IRQ_ALIGN 4

// The SW IRQ is ID 129 on AM263x. Override it if needed.
#ifndef RT_VIM_SW_IRQ
#define RT_VIM_SW_IRQ 129
#endif

// The SW IRQ must have the lowest priority.
#define RT_VIM_SW_IRQ_PRIORITY 15

#define RT_VIM 0x50F00000U
#define RT_VIM_IRQVEC_OFFSET 0x18
#define RT_VIM_IRQPRIMSK_OFFSET 0x28
#define RT_VIM_GROUPS_OFFSET 0x400

#define RT_VIM_GROUP_RAW_OFFSET 0x0
#define RT_VIM_GROUP_STS_OFFSET 0x4

#ifndef __ASSEMBLER__

#include <vim.h>

#include <rt/assert.h>

static inline void vic_syscall_pend(void)
{
    vim_set(RT_VIM_SW_IRQ);
}

rt_static_assert(VIM_IRQ_ALIGN == RT_VIC_IRQ_ALIGN,
                 "RT_VIC_IRQ_ALIGN is incorrect");
rt_static_assert((uintptr_t)VIM == RT_VIM, "RT_VIM is incorrect");
rt_static_assert((uintptr_t)&VIM->irqvec == RT_VIM + RT_VIM_IRQVEC_OFFSET,
                 "RT_VIM_IRQVEC_OFFSET is incorrect");
rt_static_assert((uintptr_t)&VIM->irqprimsk == RT_VIM + RT_VIM_IRQPRIMSK_OFFSET,
                 "RT_VIM_IRQPRIMSK_OFFSET is incorrect");
rt_static_assert((uintptr_t)&VIM->groups == RT_VIM + RT_VIM_GROUPS_OFFSET,
                 "RT_VIM_GROUPS_OFFSET is incorrect");
rt_static_assert(offsetof(struct vim_group, raw) == RT_VIM_GROUP_RAW_OFFSET,
                 "RT_VIM_GROUP_RAW_OFFSET is incorrect");
rt_static_assert(offsetof(struct vim_group, sts) == RT_VIM_GROUP_STS_OFFSET,
                 "RT_VIM_GROUP_STS_OFFSET is incorrect");

#endif // !defined(__ASSEMBLER__)
