#include <rt/context.h>
#include <rt/cycle.h>
#include <rt/idle.h>
#include <rt/interrupt.h>
#include <rt/mpu.h>
#include <rt/syscall.h>
#include <rt/task.h>
#include <rt/tls.h>
#include <rt/trap.h>

#include <rt/arch/fp.h>
#include <rt/arch/profile.h>

#include <stddef.h>
#include <stdint.h>

#if RT_ARM_PROFILE_AR
#include "ar/ar.c"
#elif RT_ARM_PROFILE_M
#include "m/m.c"
#endif

__attribute__((noreturn)) void rt_task_entry(void (*fn)(void));

__attribute__((noreturn)) void rt_task_entry_arg(uintptr_t arg,
                                                 void (*fn)(uintptr_t));

void *rt_context_init(void (*fn)(void), void *stack, size_t stack_size)
{
    struct context *const ctx = context_init(stack, stack_size, (uintptr_t)fn);
    ctx->pc = (uint32_t)rt_task_entry;
    ctx->r0 = (uint32_t)fn;
    return ctx;
}

void *rt_context_init_arg(void (*fn)(uintptr_t), uintptr_t arg, void *stack,
                          size_t stack_size)
{
    struct context *const ctx = context_init(stack, stack_size, (uintptr_t)fn);
    ctx->pc = (uint32_t)rt_task_entry_arg;
    ctx->r0 = arg;
    ctx->r1 = (uint32_t)fn;
    return ctx;
}

__attribute__((noreturn, weak)) void rt_idle(void)
{
    rt_task_drop_privilege();
    for (;;)
    {
        __asm__("wfi" :::);
    }
}

__attribute__((noreturn, weak)) void rt_trap(void)
{
    for (;;)
    {
        __asm__("bkpt" :::);
    }
}

#if RT_MPU_TASK_REGIONS_ENABLE
rt_static_assert(offsetof(struct rt_task, mpu_config) ==
                     RT_MPU_TASK_CONFIG_OFFSET,
                 "RT_MPU_TASK_CONFIG_OFFSET is incorrect");
#endif // RT_MPU_TASK_REGIONS_ENABLE

#if RT_ARM_V6M
#include "m/atomic-v6.c"
#endif
