#pragma once

#include <rt/mpu.h>

#include <rt/arch/fp.h>

#ifdef __cplusplus
extern "C" {
#endif

#if RT_MPU_TASK_REGIONS_ENABLE

#define RT_STACK_ALIGN(n) RT_MPU_ALIGN(n)

#if __ARM_ARCH == 6 || __ARM_ARCH == 7
#define RT_STACK_SIZE(n) RT_MPU_SIZE(n)
#endif

#endif // RT_MPU_TASK_REGIONS_ENABLE

// The Arm ABI specifies a minimum stack alignment of 8 bytes.
#ifndef RT_STACK_ALIGN
#define RT_STACK_ALIGN(n) 8UL
#endif

/* A register context takes 16-18 words depending on the architecture
 * subvariant. Pushing all callee-saved registers once requires an
 * additional 8 words. */
#ifndef RT_STACK_MIN
#define RT_STACK_MIN 128
#endif

/* With hardware floating-point, the context is 136-144 bytes larger: 32s/16d
 * registers, 1-3 floating-point state registers (fpscr on M and R, fpexc,
 * cpacr on R). For *d32 FPUs, add another 16*8 = 128 bytes. */
#ifndef RT_STACK_FP_MIN
#if RT_ARM_FP
#if RT_ARM_FP_D32
#define RT_STACK_FP_MIN 512
#else // !RT_ARM_FP_D32
#define RT_STACK_FP_MIN 384
#endif // RT_ARM_FP_D32
#else  // !RT_ARM_FP
#define RT_STACK_FP_MIN RT_STACK_MIN
#endif // RT_ARM_FP
#endif // RT_STACK_FP_MIN

#ifdef __cplusplus
}
#endif
