#pragma once

#include <rt/assert.h>

#include <stddef.h>

// Arm accesses tls variables with an offset of 8 bytes from the thread pointer.
#define RT_TLS_OFFSET 8

static inline void rt_tls_init(void *tls, size_t size)
{
    extern const char __tdata__[], __tdata_size__, __tbss_offset__, __tbss_size__;
    // TODO: figure out format of tls on arm
    char *const local_tdata = (char *)tls + RT_TLS_OFFSET;
    const size_t tdata_size = (size_t)&__tdata_size__;
    /* The .tbss offset is at least as large as tdata_size. It will be larger if
     * there is any alignment padding between .tdata and .tbss. */
    const size_t tbss_offset = (size_t)&__tbss_offset__;
    char *const local_tbss = local_tdata + tbss_offset;
    const size_t tbss_size = (size_t)&__tbss_size__;
    rt_assert(size >= RT_TLS_OFFSET + tbss_offset + tbss_size, "tls size is too small");
    __builtin_memcpy(local_tdata, __tdata__, tdata_size);
    __builtin_bzero(local_tbss, tbss_size);
}
