@ vim:ft=arm

/* This variant of the syscall handler supports armv6-m and armv8-m.base.
 *
 * armv8-m.base also supports cbz while armv6-m does not, but otherwise they
 * support the same subset of armv{7,8}-m that makes the save/restore procedure
 * different, mainly due to the lack of stmdb and ldm.w/stm.w. */

#include <rt/mpu.h>

#include <rt/arch/pseudo.S>

#include "exc_return.h"

    .syntax unified

#if RT_MPU_TASK_REGIONS_ENABLE
.macro mpuconfigure
    mov32 r1, rt_active_task
    ldr r1, [r1]
    adds r1, RT_MPU_TASK_CONFIG_OFFSET
#if __ARM_ARCH == 8
    /* In armv8-m, the region number isn't part of the base address
     * register and must be set separately. */
    ldr r2, =(RT_MPU_BASE + RT_MPU_REGION_NUMBER_OFFSET)
#else // __ARM_ARCH == 6
    ldr r2, =(RT_MPU_BASE + RT_MPU_REGION_OFFSET)
#endif // __ARM_ARCH
    mpuset RT_MPU_TASK_REGION_START_ID, RT_MPU_NUM_TASK_REGIONS
    dsb
.endm

.macro mpuset r, n
    .if \n == 0
    .exitm
    .endif
    ldmia r1!, {r4-r5}
#if __ARM_ARCH == 8
    mov r3, \r
    stmia r2!, {r3, r4, r5}
    subs r2, 12
#else // __ARM_ARCH == 6
    stmia r2!, {r4, r5}
    subs r2, 8
#endif // __ARM_ARCH
    mpuset (\r + 1), (\n - 1)
.endm
#else // !RT_MPU_TASK_REGIONS_ENABLE
.macro mpuconfigure
.endm
#endif // RT_MPU_TASK_REGIONS_ENABLE

#if (__ARM_ARCH >= 7) && RT_MPU_TASK_REGIONS_ENABLE
#define CONTROL_SIZE 4
#define controltemp r2,
#define getcontrol mrs r2, control
#define setcontrol msr control, r2
#else // __ARM_ARCH <= 6 || !RT_MPU_TASK_REGIONS_ENABLE
#define CONTROL_SIZE 0
#define controltemp
#define getcontrol
#define setcontrol
#endif // (__ARM_ARCH >= 7) && RT_MPU_TASK_REGIONS_ENABLE

#if __ARM_ARCH == 8
#define PSPLIM_SIZE 4
#define psplimtemp r3,
#define getpsplim mrs r3, psplim
#define setpsplim msr psplim, r3
#else // __ARM_ARCH == 6
#define PSPLIM_SIZE 0
#define psplimtemp
#define getpsplim
#define setpsplim
#endif // __ARM_ARCH

#define CONTEXT_SIZE (32 + PSPLIM_SIZE + CONTROL_SIZE)

.macro swapcontext
    // Store the suspending task's context to its stack.
    mrs r1, psp
    subs r1, CONTEXT_SIZE

    mov32 r2, rt_context_prev
    ldr r2, [r2]
    str r1, [r2]

    getcontrol
    getpsplim
    stmia r1!, {controltemp psplimtemp r4-r7}
    mov r4, r8
    mov r5, r9
    mov r6, r10
    mov r7, r11
    stmia r1!, {r4-r7}

    mpuconfigure

    // Load the new context returned by rt_syscall_run_pending.
    mov r1, r0
    adds r0, CONTEXT_SIZE - 16
    ldmia r0!, {r4-r7}
    mov r8, r4
    mov r9, r5
    mov r10, r6
    mov r11, r7
    ldmia r1!, {controltemp psplimtemp r4-r7}

    // Set the new stack pointer.
    msr psp, r0
    setcontrol
    setpsplim
.endm


    .section .text.rt_svcall_handler,"ax",%progbits
    .global rt_svcall_handler
    .type rt_svcall_handler, %function
rt_svcall_handler:
    bl rt_syscall_run
    cbz r0, 0f
    swapcontext
0:  movs r0, TASK_INITIAL_EXC_RETURN_NEG
    negs r0, r0
    bx r0

    .size rt_svcall_handler, .-rt_svcall_handler


    .section .text.rt_pendsv_handler,"ax",%progbits
    .global rt_pendsv_handler
    .type rt_pendsv_handler, %function
rt_pendsv_handler:
    bl rt_syscall_run_pending
    cbz r0, 0f
    swapcontext
0:  movs r0, TASK_INITIAL_EXC_RETURN_NEG
    negs r0, r0
    bx r0

    .size rt_pendsv_handler, .-rt_pendsv_handler


    .section .text.rt_start,"ax",%progbits
    .global rt_start
    .type rt_start, %function
rt_start:
    bl rt_start_context
    mpuconfigure
    mov r1, r0
    adds r0, CONTEXT_SIZE - 16
    ldmia r0!, {r4-r7}
    mov r8, r4
    mov r9, r5
    mov r10, r6
    mov r11, r7
    ldmia r1!, {controltemp psplimtemp r4-r7}
    msr psp, r0
    setpsplim
    // Set the SPSEL bit in control to switch to the process stack pointer.
    movs r0, 2
    msr control, r0
    isb
    // Once we are using the process stack pointer, interrupts can be enabled.
    cpsie i
    /* Manually perform an exception return stack pop. This ignores the
     * context's initial PSR, but the current PSR works as well. We can't pop
     * all of the registers at once in v6/v8-m.base, so we need to load r12 and
     * lr individually and copy pc to the psr slot so it can be popped. */
    ldr r0, [sp, 16]
    mov r12, r0
    ldr r0, [sp, 20]
    mov lr, r0
    ldr r0, [sp, 24]
    str r0, [sp, 28]
    pop {r0-r3}
    add sp, 12
    pop {pc}

    .size rt_start, .-rt_start
