#pragma once

// The C28x ABI specifies a minimum stack alignment of 2 words.
#define RT_STACK_ALIGN(n) 2UL

#define RT_STACK_SECTION(name) ".rt_task_stack:" #name

/*
 * The integer register context takes 32 words. With FPU32 support, an
 * additional 20 words are needed, and with FPU64 support, an additional 16
 * words on top of that. Pushing all callee-saved registers once requires an
 * additional 8 words. The system call handler will run on each task stack as
 * well.
 */
#ifdef __TMS320C28XX_FPU64__

// The double-precision floating-point context is an additional 36 words.
#define RT_STACK_MIN 256UL

#elif defined(__TMS320C28XX_FPU32__)

// The floating-point context is an additional 20 words.
#define RT_STACK_MIN 192UL

#else // !defined(__TMS320C28XX_FPU64__) && !defined(__TMS320C28XX_FPU32__)

#define RT_STACK_MIN 128UL

#endif

/* The c28 context switch always saves and restores floating-point context even
 * if a task doesn't use it, so there's no separate FP stack minimum. */
#define RT_STACK_FP_MIN RT_STACK_MIN
