#pragma once

#include <rt/mutex.h>

namespace rt
{

class mutex
{
  public:
    constexpr mutex() noexcept : m(RT_MUTEX_INIT(m))
    {
    }

    void lock() noexcept
    {
        rt_mutex_lock(&this->m);
    }

    void unlock() noexcept
    {
        rt_mutex_unlock(&this->m);
    }

    bool trylock() noexcept
    {
        return rt_mutex_trylock(&this->m);
    }

    bool timedlock(unsigned long ticks) noexcept
    {
        return rt_mutex_timedlock(&this->m, ticks);
    }

    friend class cond;

    mutex(const mutex &) = delete;
    mutex(const mutex &&) = delete;
    mutex &operator=(const mutex &) = delete;
    mutex &operator=(const mutex &&) = delete;

  protected:
    struct is_recursive_t
    {
        explicit is_recursive_t() noexcept = default;
    };
    static inline constexpr is_recursive_t is_recursive{};

    constexpr mutex(is_recursive_t) noexcept : m(RT_MUTEX_INIT_RECURSIVE(m))
    {
    }

  private:
    struct rt_mutex m;
};

class recursive_mutex : public mutex
{
  public:
    constexpr recursive_mutex() noexcept : mutex(is_recursive)
    {
    }
};

struct adopt_lock_t
{
    explicit adopt_lock_t() noexcept = default;
};

inline constexpr adopt_lock_t adopt_lock{};

class lock_guard
{
  public:
    explicit lock_guard(mutex &m) noexcept : mtx(m)
    {
        mtx.lock();
    }

    lock_guard(mutex &m, adopt_lock_t) noexcept : mtx(m)
    {
    }

    ~lock_guard() noexcept
    {
        mtx.unlock();
    }

    lock_guard(const lock_guard &) = delete;
    lock_guard &operator=(const lock_guard &) = delete;

  private:
    mutex &mtx;
};

} // namespace rt
