#pragma once

#include <rt/notify.h>

namespace rt
{

class notify
{
  public:
    constexpr notify(uint32_t value = 0) noexcept : n(RT_NOTIFY_INIT(n, value))
    {
    }

    void nop() noexcept
    {
        rt_notify_post(&this->n);
    }

    void orr(uint32_t value) noexcept
    {
        rt_notify_or(&this->n, value);
    }

    void add(uint32_t value) noexcept
    {
        rt_notify_add(&this->n, value);
    }

    void set(uint32_t value) noexcept
    {
        rt_notify_set(&this->n, value);
    }

    uint32_t wait() noexcept
    {
        return rt_notify_wait(&this->n);
    }

    uint32_t wait_clear(uint32_t clear) noexcept
    {
        return rt_notify_wait_clear(&this->n, clear);
    }

    bool trywait(uint32_t &value) noexcept
    {
        return rt_notify_trywait(&this->n, &value);
    }

    bool trywait_clear(uint32_t &value, uint32_t clear) noexcept
    {
        return rt_notify_trywait_clear(&this->n, &value, clear);
    }

    bool timedwait(uint32_t &value, unsigned long ticks) noexcept
    {
        return rt_notify_timedwait(&this->n, &value, ticks);
    }

    bool timedwait_clear(uint32_t &value, uint32_t clear,
                         unsigned long ticks) noexcept
    {
        return rt_notify_timedwait_clear(&this->n, &value, clear, ticks);
    }

    notify(const notify &) = delete;
    notify(const notify &&) = delete;
    notify &operator=(const notify &) = delete;
    notify &operator=(const notify &&) = delete;

  private:
    struct rt_notify n;
};

} // namespace rt
