#pragma once

#include <rt/queue.h>

namespace rt
{

template <typename T, size_t num> class queue
{
  public:
    constexpr queue() noexcept : q(RT_QUEUE_INIT(q, T, num, slots, data))
    {
    }

    void push(const T &e) noexcept
    {
        rt_queue_push(&this->q, &e);
    }

    void pop(T &e) noexcept
    {
        rt_queue_pop(&this->q, &e);
    }

    void peek(T &e) noexcept
    {
        rt_queue_peek(&this->q, &e);
    }

    bool trypush(const T &e) noexcept
    {
        return rt_queue_trypush(&this->q, &e);
    }

    bool trypop(T &e) noexcept
    {
        return rt_queue_trypop(&this->q, &e);
    }

    bool trypeek(T &e) noexcept
    {
        return rt_queue_trypeek(&this->q, &e);
    }

    bool timedpush(const T &e, unsigned long ticks) noexcept
    {
        return rt_queue_timedpush(&this->q, &e, ticks);
    }

    bool timedpop(T &e, unsigned long ticks) noexcept
    {
        return rt_queue_timedpop(&this->q, &e, ticks);
    }

    bool timedpeek(T &e, unsigned long ticks) noexcept
    {
        return rt_queue_timedpeek(&this->q, &e, ticks);
    }

    queue(const queue &) = delete;
    queue(const queue &&) = delete;
    queue &operator=(const queue &) = delete;
    queue &operator=(const queue &&) = delete;

  private:
    static_assert(num <= RT_QUEUE_MAX_SIZE, "queue is too large");
    struct rt_queue q;
    rt_atomic_uchar slots[num];
    unsigned char data[num * sizeof(T)];
};

} // namespace rt
