#pragma once

#include <rt/rwlock.h>

#include <rt/mutex.hpp>

namespace rt
{

class rwlock
{
  public:
    constexpr rwlock() noexcept : l(RT_RWLOCK_INIT(l))
    {
    }

    void rdlock() noexcept
    {
        rt_rwlock_rdlock(&this->l);
    }

    void wrlock() noexcept
    {
        rt_rwlock_wrlock(&this->l);
    }

    void unlock() noexcept
    {
        rt_rwlock_unlock(&this->l);
    }

    rwlock(const rwlock &) = delete;
    rwlock(const rwlock &&) = delete;
    rwlock &operator=(const rwlock &) = delete;
    rwlock &operator=(const rwlock &&) = delete;

  private:
    struct rt_rwlock l;
};

class rdlock_guard
{
  public:
    explicit rdlock_guard(rwlock &l) noexcept : rwl(l)
    {
        rwl.rdlock();
    }

    rdlock_guard(rwlock &l, adopt_lock_t) noexcept : rwl(l)
    {
    }

    ~rdlock_guard() noexcept
    {
        rwl.unlock();
    }

    rdlock_guard(const rdlock_guard &) = delete;
    rdlock_guard &operator=(const rdlock_guard &) = delete;

  private:
    rwlock &rwl;
};

class wrlock_guard
{
  public:
    explicit wrlock_guard(rwlock &l) noexcept : rwl(l)
    {
        rwl.wrlock();
    }

    wrlock_guard(rwlock &l, adopt_lock_t) noexcept : rwl(l)
    {
    }

    ~wrlock_guard() noexcept
    {
        rwl.unlock();
    }

    wrlock_guard(const wrlock_guard &) = delete;
    wrlock_guard &operator=(const wrlock_guard &) = delete;

  private:
    rwlock &rwl;
};

} // namespace rt
