#pragma once

#include <rt/sem.h>

namespace rt
{

class sem
{
  public:
    constexpr sem(int count = 0, int max = INT_MAX) noexcept
        : s(RT_SEM_INIT_MAX(s, count, max))
    {
    }

    void post(int n = 1) noexcept
    {
        rt_sem_post_n(&this->s, n);
    }

    void wait() noexcept
    {
        rt_sem_wait(&this->s);
    }

    bool trywait() noexcept
    {
        return rt_sem_trywait(&this->s);
    }

    bool timedwait(unsigned long ticks) noexcept
    {
        return rt_sem_timedwait(&this->s, ticks);
    }

    sem(const sem &) = delete;
    sem(const sem &&) = delete;
    sem &operator=(const sem &) = delete;
    sem &operator=(const sem &&) = delete;

  private:
    struct rt_sem s;
};

class binary_sem : public sem
{
  public:
    constexpr binary_sem(int count = 0) noexcept : sem(count, 1)
    {
    }
};

} // namespace rt
